c=======================================================================
c
      subroutine lanczos(nlan,nlanex,lfail,iresta,nlanre,nnorm,nscal,
     *                   ierf02,ht,lbloc,nbloc,eigval,eeps,
     *                   codiag,vecf02,lanmax,
     *                   eigvec,w,veclan,alpha,beta,b)
c
c=======================================================================
c     this subroutine uses the  l a n c z o s  algorithm to compute the
c     lowest part of the spectrum of a large real symmetric matrix "ht".
c     it computes  e i g e n v e c t o r s   and   e i g e n v a l u e s
c     without regard to multiplicities. Here, the full matrix "ht" is
c     needed ( ht: lbloc x lbloc ).
c     the subroutine calls :
c        -  x02aaf  function   (nag.library) .
c        -  f02amf  subroutine (nag.library)
c                   calculates the eigenvalues and the eigenvectors of
c                   a real symmetric tridiagonal matrix .
c     ==================================================================
c     arguments which must be set by the user and are not changed
c     by the subroutine .
c        - nlan   : max number of required lanczos steps .
c	          : (stop sooner if lowest eigenvalue consistent to 'eeps'
c        - nlanex : number of exact lowest eigensolutions of matrix "ht"
c          Note   : for "ht" (3432 x 3432), nlan = 60 ==> nlanex = 2,
c                 : with precision 10**(-9) and 10**(-5) in the two
c                 : lowest eigenvalues and 10**(-4) in the expectation
c                 : values calculated with the lowest eigenvector.
c        - lbloc  : the actual dimension of matrix "ht" .
c     ==================================================================
c     arguments which must be set by the user and are changed by the
c     subroutine .
c        - lfail  : error indicator for the subroutine "lanczos" .
c                   before entry lfail must be set to 0 .
c                   unless the routine detects an error , lfail
c                   contains 0 on exit .
c        - iresta : restart counter for the subroutine "lanczos" .
c                   before entry iresta must be set to 0 .
c                   on exit , iresta contains the total number of
c                   restart .
c        - nlanre : indicates the lanczos step where the first restart
c                   has been detected .
c                   before entry nlanre must be set to 0 .
c                   on exit , if nlanre = 0 , no restart detected .
c        - nnorm  : indicates the number of lanczos vectors with
c                   (norm-1) .gt. 1.d-14 .
c                   before entry nnorm must be set to 0 .
c        - nscal  : indicates the number of scalar products between
c                   lanczos vectors .gt. 1.d-14 .
c                   before entry nscal must be set to 0 .
c        - ierf02 : error indicator for the subroutine "f02amf" .
c                   before entry ierf02 must be set to 99 .
c                   unless the routine detects an error , ierf02
c                   contains 0 on exit .
c     ==================================================================
c     arguments which provide results but need not be set by the user .
c        - eigval : vector of dimension (lanmax) .
c                   on return , it contains nlan computed eigenvalues
c                   stored in increasing order .only the first nlanex
c                   eigenvalues are available .
c     ==================================================================
c     dummy arguments .
c        - codiag : vector of dimension (lanmax) .
c                   it contains the sub-diagonal of the real symmetric
c                   tridiagonal matrix to diagonalize .
c        - vecf02 : matrix of dimension (lanmax,lanmax) .
c                   it contains the eigenvectors of the real symmetric
c                   tridiagonal matrix to diagonalize .
c     ==================================================================
c     output :
c        - eigvec : matrix of dimension (nbloc,nlanex) .
c                   on successful exit , the matrix contains the norma-
c                   lised eigenvectors such that eigvec(i,j) i=1,lbloc
c                   corresponds to eigenvalue eigval(j) .
c=======================================================================
c
      implicit integer
     *                (i-n,p)
c
      implicit double precision
     *                         (a-h,o,q-z)
c
      dimension
     *         eigval(lanmax),codiag(lanmax),vecf02(lanmax,lanmax),
     *            ht(nbloc,nbloc),eigvec(nbloc,nlanex), w(nbloc),
     *            veclan(nbloc,lanmax),alpha(lanmax),beta(lanmax-1),
     *            b(lanmax)
c
c      write(6,*) ' Lanczos: nlan,nlanex,lbloc,nbloc,eeps,lanmax =',
c     x			nlan,nlanex,lbloc,nbloc,eeps,lanmax

      if (nlan.le.0) then
         lfail=1
         go to 111
      end if
c
      if (nlan.gt.lbloc) then
         lfail=2
         go to 111
      end if
c
      if (nlanex.gt.nlan) then
         lfail=5
         go to 111
      end if
c
      eps=x02ajf(1.d+00)
c
c     generating pseudo-random normalized start vector
c     ------------------------------------------------
c
      vlvl=0.d+00
      ggg0=1431655765.d+00
c
      do i=1,lbloc
         ggg0=mod(ggg0*9228907.d+00,4294967296.d+00)
         veclan(i,1) = 2.d+00*ggg0/4294967296.d+00 - 1.d+00
         vlvl = vlvl + veclan(i,1)*veclan(i,1)
      end do
c
      vlvl=1.d+00/sqrt(vlvl)
c
      do i=1,lbloc
         veclan(i,1) = veclan(i,1)*vlvl
      end do
c
c     performing the first lanczos step
c     ---------------------------------
c
      alpha(1)=0.d+00
      gamma=0.d+00
      lstop1=0
c
      do i=1,lbloc
         w(i)=0.d+00
         do j=1,lbloc
            w(i)=w(i)+ht(j,i)*veclan(j,1)
         end do
         if (abs(w(i)).le.1.d-06) lstop1=lstop1+1
         alpha(1)=alpha(1)+veclan(i,1)*w(i)
         gamma=gamma+w(i)*w(i)
      end do
c
      if (lstop1.eq.lbloc) then
         lfail=3
         go to 111
      end if
c
      beta(1) = sqrt( gamma-alpha(1)*alpha(1) )
c
      if (abs(beta(1)).le.1.d-06) then
         lfail=4
         go to 111
      end if
c
      do i=1,lbloc
         veclan(i,2)= (w(i)-alpha(1)*veclan(i,1))/beta(1)
      end do
c
      write(10,'(10x,'' lanczos step : '',1i3,2x,''done'')') 1
	elast = 1d6
c
c     performing lanczos steps : n (2) ---> n (nlan-1)
c     --------------------------------------------------
c
      do l=2,nlan-1
c
         alpha(l)=0.d+00
c
         do i=1,lbloc
            w(i)=0.d+00
            do j=1,lbloc
               w(i)=w(i)+ht(j,i)*veclan(j,l)
            end do
            alpha(l) = alpha(l) + veclan(i,l)*w(i)
         end do
c
         do i=1,nlan
            b(i)=0.d+00
         end do
c
         do i=1,lbloc
            w(i)=w(i)-beta(l-1)*veclan(i,l-1)-alpha(l)*veclan(i,l)
            do m=1,l
               b(m)=b(m)+veclan(i,m)*w(i)
            end do
         end do
c
         delta=0.d+00
c
         do i=1,lbloc
            do m=1,l
               w(i)=w(i)-b(m)*veclan(i,m)
            end do
            delta=delta+w(i)*w(i)
         end do
c
         beta(l)=sqrt(delta)
c
         if (abs(beta(l)).le.1.d-06) then
c
            beta(l)=0.d+00
            iresta=iresta+1
c
            if (iresta.eq.1) nlanre=l
c
            write(10,'(25x,''--> restart occured '')')
c
c           generating pseudo-random normalized restart vector
c           --------------------------------------------------
c
            vlvl=0.d+00
            ggg1=1431655765.d+00*(iresta+1)
c
            do i=1,lbloc
               ggg1=mod(ggg1*9228907.d+00,4294967296.d+00)
               w(i) = 2.d+00*ggg1/4294967296.d+00 - 1.d+00
               vlvl = vlvl + w(i)*w(i)
            end do
c
            vlvl=1.d+00/sqrt(vlvl)
c
            do i=1,lbloc
               w(i)=w(i)*vlvl
            end do
c
            do i=1,nlan
               b(i)=0.d+00
            end do
c
            do i=1,lbloc
               do m=1,l
                  b(m)=b(m)+veclan(i,m)*w(i)
               end do
            end do
c
            delta=0.d+00
c
            do i=1,lbloc
               do m=1,l
                  w(i)=w(i)-b(m)*veclan(i,m)
               end do
               delta=delta+w(i)*w(i)
            end do
c
            s=1.d+00/sqrt(delta)
c
            do i=1,lbloc
               veclan(i,l+1)=w(i)*s
            end do
c
         else
c
            s=1.d+00/sqrt(delta)
c
            do i=1,lbloc
               veclan(i,l+1)=w(i)*s
            end do
c
         end if
c
         write(10,'(10x,''         step : '',1i3,2x,''done'')') l
c
	if(mod(l,nlan/10).eq.0.and.l.gt.nlan/3) then
c        diagonalizing partial lanczos tridiagonal matrix ( nag : "f02amf" )
c        -----------------------------------------------------------
c
      do i=1,l-1
         eigval(i)=alpha(i)
         codiag(i+1)=beta(i)
      end do
c
      codiag(1)=10.d+00
      eigval(l)=alpha(l)
c
      do i=1,l
         do j=1,l
            vecf02(j,i)=0.d+00
         end do
         vecf02(i,i)=1.d+00
      end do
c
      call f02amf(l,eps,eigval,codiag,vecf02,lanmax,ierf02)
	write(6,99) l,(eigval(i),i=1,min(10,nlanex))
	write(10,99) l,(eigval(i),i=1,min(10,nlanex))
	if(abs(eigval(1)-elast).lt.eeps) go to 100
	elast = eigval(1)	
	endif
      end do
c
c     performing the last lanczos step
c     --------------------------------
c
      alpha(nlan)=0.d+00
c
      do i=1,lbloc
         w(i)=0.d+00
         do j=1,lbloc
            w(i)=w(i)+ht(j,i)*veclan(j,nlan)
         end do
         alpha(nlan) = alpha(nlan)+veclan(i,nlan)*w(i)
      end do
c
      write(10,'(10x,''  last   step : '',1i3,2x,''done'')') nlan
c
c     testing orthonormalization
c     --------------------------
c
      do m=1,nlan
         do l=m,nlan
            scal=0.d+00
            do j=1,lbloc
               scal=scal+veclan(j,m)*veclan(j,l)
            end do
            if (m.eq.l) then
               if (abs(scal-1.d+00).gt.1.d-14) nnorm=nnorm+1
            else
               if (abs(scal).gt.1.d-14) nscal=nscal+1
            end if
         end do
      end do
c
c     diagonalizing lanczos tridiagonal matrix ( nag : "f02amf" )
c     -----------------------------------------------------------
c
      do i=1,nlan-1
         eigval(i)=alpha(i)
         codiag(i+1)=beta(i)
      end do
c
      codiag(1)=10.d+00
      eigval(nlan)=alpha(nlan)
c
      do i=1,nlan
         do j=1,nlan
            vecf02(j,i)=0.d+00
         end do
         vecf02(i,i)=1.d+00
      end do
c
      call f02amf(nlan,eps,eigval,codiag,vecf02,lanmax,ierf02)
	write(10,99) nlan,(eigval(i),i=1,min(10,nlanex))
99	format(' After ',i4,', lanczos eigenvalues =',10f10.6)
c
      if (ierf02.eq.1) go to 111
c
c     performing eigenvectors in the original basis
c     ---------------------------------------------
c
100    do i=1,lbloc
         do j=1,nlanex
            s=0.d+00
            do l=1,nlan
               s=s+veclan(i,l)*vecf02(l,j)
            end do
            eigvec(i,j)=s
         end do
      end do
c
 111  continue
      return
c
      end

